import pkgutil
from collections.abc import Collection, Iterator, Mapping, Sequence
from contextlib import ContextDecorator, suppress
from functools import cached_property, singledispatchmethod
from pathlib import Path
from typing import Annotated, Any, Optional
from warnings import warn

from filelock import FileLock
from pydantic import (
    BaseModel,
    Field,
    TypeAdapter,
    ValidationInfo,
    field_validator,
    model_validator,
)
from typer import get_app_dir

__version__ = "0.4.6"


class Constituent(BaseModel, populate_by_name=True, frozen=True):
    id: Optional[int] = None
    name: str
    u_neg: Annotated[
        list[float], Field(alias="uNeg")
    ] = []  # [-neg_count, -neg_count+1, -neg_count+2, ..., -1]
    u_pos: Annotated[
        list[float], Field(alias="uPos")
    ] = []  # [+1, +2, +3, ..., +pos_count]
    pkas_neg: Annotated[
        list[float], Field(alias="pKaNeg")
    ] = []  # [-neg_count, -neg_count+1, -neg_count+2, ..., -1]
    pkas_pos: Annotated[
        list[float], Field(alias="pKaPos")
    ] = []  # [+1, +2, +3, ..., +pos_count]
    neg_count: Annotated[int, Field(alias="negCount", validate_default=True)] = None  # type: ignore
    pos_count: Annotated[int, Field(alias="posCount", validate_default=True)] = None  # type: ignore

    def mobilities(self) -> Sequence[float]:
        n = max(self.neg_count, self.pos_count, 3)
        ret = (
            [0.0] * (n - self.pos_count)
            + [u * 1e-9 for u in self.u_pos[::-1]]
            + [u * 1e-9 for u in self.u_neg[::-1]]
            + [0.0] * (n - self.neg_count)
        )
        assert len(ret) == 2 * n
        return ret

    def pkas(self) -> Sequence[float]:
        n = max(self.neg_count, self.pos_count, 3)
        ret = (
            [self._default_pka(c) for c in range(n, self.pos_count, -1)]
            + self.pkas_pos[::-1]
            + self.pkas_neg[::-1]
            + [self._default_pka(-c) for c in range(self.neg_count + 1, n + 1)]
        )
        assert len(ret) == 2 * n
        return ret

    def diffusivity(self) -> float:
        mobs = []
        with suppress(IndexError):
            mobs.append(self.u_neg[-1] * 1e-9)
        with suppress(IndexError):
            mobs.append(self.u_pos[0] * 1e-9)

        return max(mobs, default=0) * 8.314 * 300 / 96485

    @staticmethod
    def _default_pka(charge: int) -> float:
        assert charge != 0
        if charge < 0:
            return 14 - charge
        return -charge

    @field_validator("name")
    def _normalize_db1_names(cls, v: str, info: ValidationInfo) -> str:
        if info.context and info.context.get("fix", None) == "db1":
            v = v.replace(" ", "_").replace("Cl-", "CHLORO")
        return v

    @field_validator("name")
    def _no_whitespace(cls, v: str, info: ValidationInfo) -> str:
        parts = v.split()
        if len(parts) > 1 or len(parts[0]) != len(v):
            raise ValueError("name cannot contain any whitespace")
        return parts[0]

    @field_validator("name")
    def _all_uppercase(cls, v: str, info: ValidationInfo) -> str:
        if not v.isupper():
            raise ValueError("name must be all uppercase")
        return v

    @field_validator("pkas_neg", "pkas_pos")
    def _pka_lengths(cls, v: list[float], info: ValidationInfo) -> list[float]:
        assert isinstance(info.field_name, str)
        if len(v) != len(info.data[f"u_{info.field_name[5:]}"]):
            raise ValueError(f"len({info.field_name}) != len(u_{info.field_name[5:]})")
        return v

    @field_validator("neg_count", "pos_count", mode="before")
    def _counts(cls, v: Optional[int], info: ValidationInfo) -> int:
        assert isinstance(info.field_name, str)
        if v is None:
            v = len(info.data[f"u_{info.field_name[:3]}"])
        elif v != len(info.data[f"u_{info.field_name[:3]}"]):
            raise ValueError(f"{info.field_name} != len(u_{info.field_name[:3]})")
        return v

    @model_validator(mode="after")
    def _pkas_not_increasing(self) -> "Constituent":
        pkas = [*self.pkas_neg, *self.pkas_pos]

        if not all(x >= y for x, y in zip(pkas, pkas[1:])):
            raise ValueError("pKa values must not increase with charge")

        return self


_StoredConstituents = TypeAdapter(dict[str, list[Constituent]])


def _load_constituents(
    data: bytes, context: Optional[dict[str, str]] = None
) -> list[Constituent]:
    return _StoredConstituents.validate_json(data, context=context)["constituents"]


def _dump_constituents(constituents: list[Constituent]) -> bytes:
    return _StoredConstituents.dump_json(
        {"constituents": constituents}, by_alias=True, indent=4
    )


class _Database(Mapping[str, Constituent], ContextDecorator):
    def __init__(self, user_constituents_file: Path) -> None:
        self._user_constituents_file = user_constituents_file
        self._user_constituents_lock = FileLock(
            self._user_constituents_file.with_suffix(".lock")
        )
        self._user_constituents_dirty = False

    @cached_property
    def _default_constituents(self) -> dict[str, Constituent]:
        data = pkgutil.get_data(__package__, "db1.json")
        if data is None:
            raise RuntimeError("failed to load default constituents")
        constituents = _load_constituents(data, context={"fix": "db1"})
        return {c.name: c for c in constituents}

    @cached_property
    def _user_constituents(self) -> dict[str, Constituent]:
        try:
            with self:
                user_data = self._user_constituents_file.read_bytes()
        except OSError:
            return {}
        try:
            user_constituents = _load_constituents(user_data)
        except Exception as e:
            warn(
                f"failed to load user constituents from {self._user_constituents_file}: {type(e).__name__}",
                RuntimeWarning,
                stacklevel=2,
            )
            return {}
        return {c.name: c for c in user_constituents}

    def _invalidate_user_constituents(self) -> None:
        assert not self._user_constituents_dirty
        with suppress(AttributeError):
            del self._user_constituents

    def _save_user_constituents(self) -> None:
        data = _dump_constituents(list(self._user_constituents.values()))
        self._user_constituents_file.parent.mkdir(parents=True, exist_ok=True)
        with self:
            self._user_constituents_file.write_bytes(data)
        self._user_constituents_dirty = False

    def __enter__(self) -> "_Database":
        if not self._user_constituents_lock.is_locked:
            self._invalidate_user_constituents()
        self._user_constituents_lock.acquire()

        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        try:
            if (
                self._user_constituents_lock.lock_counter == 1
                and self._user_constituents_dirty
            ):
                self._save_user_constituents()
        finally:
            self._user_constituents_lock.release()

    def __iter__(self) -> Iterator[str]:
        yield from sorted([*self._default_constituents, *self._user_constituents])

    def __getitem__(self, name: str) -> Constituent:
        name = name.upper()
        try:
            return self._user_constituents[name]
        except KeyError:
            return self._default_constituents[name]

    def add(self, constituent: Constituent) -> None:
        with self:
            if constituent.name not in self:
                self._user_constituents[constituent.name] = constituent
                self._user_constituents_dirty = True
            else:
                warn(
                    f"{constituent.name}: component was not added (name already exists in database)",
                    stacklevel=2,
                )

    def __delitem__(self, name: str) -> None:
        name = name.upper()
        try:
            with self:
                del self._user_constituents[name]
                self._user_constituents_dirty = True
        except KeyError:
            if name in self._default_constituents:
                raise ValueError(f"{name}: cannot remove default component") from None
            raise

    def __len__(self) -> int:
        return len(self._default_constituents) + len(self._user_constituents)

    def user_defined(self) -> Collection[str]:
        return sorted(self._user_constituents)

    @singledispatchmethod
    def __contains__(self, _: Any) -> bool:  # type: ignore
        return False

    @__contains__.register
    def _(self, obj: str) -> bool:
        name = obj.upper()
        return name in self._default_constituents or name in self._user_constituents

    @__contains__.register
    def _(self, obj: Constituent) -> bool:
        try:
            return obj == self[obj.name]
        except KeyError:
            return False

    def is_user_defined(self, name: str) -> bool:
        name = name.upper()
        return name in self._user_constituents


database = _Database(Path(get_app_dir(__package__), "user_constituents.json"))



import pytest

import electrolytes
from electrolytes import Constituent, database


def test_version() -> None:
    assert isinstance(electrolytes.__version__, str)


def test_list_components() -> None:
    assert database
    lst = list(database)
    assert len(lst) == len(database)
    assert isinstance(lst[0], str)
    assert lst == sorted(lst)
    assert "LYSINE" in lst
    assert "CYSTINE" in lst
    assert "SILVER" in lst


def test_get_component() -> None:
    c = database["LYSINE"]
    assert isinstance(c, Constituent)
    assert c.name == "LYSINE"
    assert c.name in database
    assert c in database  # type: ignore [comparison-overlap]
    assert c.name not in database.user_defined()
    assert c not in database.user_defined()  # type: ignore [comparison-overlap]
    assert len(c.mobilities()) == 6
    assert len(c.pkas()) == 6
    assert c.diffusivity() == pytest.approx(28.60 * 1e-9 * 8.314 * 300 / 96485)


def test_known_component_properties() -> None:
    c = database["CYSTINE"]
    assert c.mobilities() == pytest.approx(
        [0.0, 5.39e-08, 2.7e-08, 2.7e-08, 5.39e-08, 0.0]
    )
    assert c.diffusivity() == pytest.approx(6.9797e-10)
    assert c.pkas() == pytest.approx([-3.0, 1.65, 2.26, 8.405, 9.845, 17])


def test_try_get_nonexistent() -> None:
    with pytest.raises(KeyError):
        database["NONEXISTENT2424612644"]


def test_try_del_nonexistent() -> None:
    with pytest.raises(KeyError):
        del database["NONEXISTENT2424612644"]


def test_try_add_default() -> None:
    assert "SILVER" in database
    assert not database.is_user_defined("SILVER")
    assert "SILVER" not in database.user_defined()
    with pytest.warns(UserWarning):
        database.add(Constituent(name="SILVER", u_pos=[64.50], pkas_pos=[11.70]))
    assert "SILVER" in database
    assert not database.is_user_defined("SILVER")
    assert "SILVER" not in database.user_defined()



from typing import Annotated, Optional

import typer

from . import Constituent, __version__, database

app = typer.Typer()


def complete_name(incomplete: str) -> list[str]:
    return [name for name in database if name.startswith(incomplete.upper())]


def complete_name_user_defined(incomplete: str) -> list[str]:
    return [
        name for name in database.user_defined() if name.startswith(incomplete.upper())
    ]


@app.command()
def add(
    name: Annotated[str, typer.Argument(autocompletion=complete_name_user_defined)],
    p1: Annotated[
        tuple[float, float],
        typer.Option("+1", help="Mobility (*1e-9) and pKa for +1", show_default=False),
    ] = (None, None),  # type: ignore
    p2: Annotated[
        tuple[float, float],
        typer.Option("+2", help="Mobility (*1e-9) and pKa for +2", show_default=False),
    ] = (None, None),  # type: ignore
    p3: Annotated[
        tuple[float, float],
        typer.Option("+3", help="Mobility (*1e-9) and pKa for +3", show_default=False),
    ] = (None, None),  # type: ignore
    p4: Annotated[
        tuple[float, float],
        typer.Option("+4", help="Mobility (*1e-9) and pKa for +4", show_default=False),
    ] = (None, None),  # type: ignore
    p5: Annotated[
        tuple[float, float],
        typer.Option("+5", help="Mobility (*1e-9) and pKa for +5", show_default=False),
    ] = (None, None),  # type: ignore
    p6: Annotated[
        tuple[float, float],
        typer.Option("+6", help="Mobility (*1e-9) and pKa for +6", show_default=False),
    ] = (None, None),  # type: ignore
    m1: Annotated[
        tuple[float, float],
        typer.Option("-1", help="Mobility (*1e-9) and pKa for -1", show_default=False),
    ] = (None, None),  # type: ignore
    m2: Annotated[
        tuple[float, float],
        typer.Option("-2", help="Mobility (*1e-9) and pKa for -2", show_default=False),
    ] = (None, None),  # type: ignore
    m3: Annotated[
        tuple[float, float],
        typer.Option("-3", help="Mobility (*1e-9) and pKa for -3", show_default=False),
    ] = (None, None),  # type: ignore
    m4: Annotated[
        tuple[float, float],
        typer.Option("-4", help="Mobility (*1e-9) and pKa for -4", show_default=False),
    ] = (None, None),  # type: ignore
    m5: Annotated[
        tuple[float, float],
        typer.Option("-5", help="Mobility (*1e-9) and pKa for -5", show_default=False),
    ] = (None, None),  # type: ignore
    m6: Annotated[
        tuple[float, float],
        typer.Option("-6", help="Mobility (*1e-9) and pKa for -6", show_default=False),
    ] = (None, None),  # type: ignore
    force: Annotated[
        bool,
        typer.Option(
            "-f",
            help="Do not prompt before replacing a user-defined component with the same name",
        ),
    ] = False,
) -> None:
    """Store a user-defined component in the database."""
    name = name.upper()

    if p1[0] is None and m1[0] is None:
        assert p1[1] is None
        assert m1[1] is None
        typer.echo("Error: at least one of the +1 or -1 options is required", err=True)
        raise typer.Exit(code=1)

    neg: list[tuple[float, float]] = []
    any_omitted = False
    for i, m in enumerate([m1, m2, m3, m4, m5, m6]):
        if m[0] is None:
            assert m[1] is None
            any_omitted = True
        elif any_omitted:
            typer.echo(f"Error: missing charge -{i}", err=True)
            raise typer.Exit(code=1)
        else:
            neg.insert(0, m)

    pos: list[tuple[float, float]] = []
    any_omitted = False
    for i, p in enumerate([p1, p2, p3, p4, p5, p6]):
        if p[0] is None:
            assert p[1] is None
            any_omitted = True
        elif any_omitted:
            typer.echo(f"Error: missing charge +{i}", err=True)
            raise typer.Exit(code=1)
        else:
            pos.append(p)

    constituent = Constituent(
        name=name,
        u_neg=[x[0] for x in neg],
        u_pos=[x[0] for x in pos],
        pkas_neg=[x[1] for x in neg],
        pkas_pos=[x[1] for x in pos],
    )

    with database:
        if name in database:
            if not database.is_user_defined(name):
                typer.echo(f"Error: {name}: is a default component", err=True)
                raise typer.Exit(code=1)

            if not force:
                typer.confirm(f"Replace existing {name}?", abort=True)

            del database[name]

        database.add(constituent)


@app.command()
def info(
    names: Annotated[
        Optional[list[str]],
        typer.Argument(help="Component names", autocompletion=complete_name),
    ] = None,
) -> None:
    """
    Show the properties of components.

    If no names are given, print the number of components in the database.
    """
    if names:
        first = True
        errors_ocurred = False
        for name in names:
            uppercase_name = name.upper()

            try:
                constituent = database[uppercase_name]
            except KeyError:
                typer.echo(f"Error: {uppercase_name}: no such component", err=True)
                errors_ocurred = True
                continue

            charges = list(range(constituent.pos_count, 0, -1)) + list(
                range(-1, -constituent.neg_count - 1, -1)
            )
            uu = constituent.u_pos[::-1] + constituent.u_neg[::-1]
            pkas = constituent.pkas_pos[::-1] + constituent.pkas_neg[::-1]

            assert len(charges) == len(uu) == len(pkas)

            if not first:
                typer.echo()
            typer.echo(f"Component: {uppercase_name}")
            if database.is_user_defined(uppercase_name):
                typer.echo("[user-defined]")
            typer.echo("                    " + " ".join(f"{c:^+8d}" for c in charges))
            typer.echo("Mobilities (*1e-9): " + " ".join(f"{u:^8.2f}" for u in uu))
            typer.echo("pKas:               " + " ".join(f"{p:^8.2f}" for p in pkas))
            typer.echo(f"Diffusivity: {constituent.diffusivity():.4e}")

            first = False

        if errors_ocurred:
            raise typer.Exit(code=1)

    else:
        total = len(database)
        user = len(database.user_defined())
        typer.echo(
            f"{total} components stored in the database ({total - user} default, {user} user-defined)"
        )


@app.command()
def ls(
    user: Annotated[
        Optional[bool],
        typer.Option(
            "--user/--default", help="List only user-defined/default components"
        ),
    ] = None,
) -> None:
    """List components in the database."""

    if user:
        names = database.user_defined()
    elif user is not None:
        names = [name for name in database if not database.is_user_defined(name)]
    else:
        names = database

    for name in names:
        typer.echo(name)


@app.command()
def rm(
    names: Annotated[
        list[str], typer.Argument(autocompletion=complete_name_user_defined)
    ],
    force: Annotated[
        Optional[bool], typer.Option("-f", help="Ignore non-existent components")
    ] = False,
) -> None:
    """Remove user-defined components from the database."""
    errors_ocurred = False
    for name in names:
        uppercase_name = name.upper()
        try:
            del database[uppercase_name]
        except KeyError:
            if not force:
                typer.echo(f"Error: {uppercase_name}: no such component", err=True)
                errors_ocurred = True
        except ValueError as e:
            typer.echo(f"Error: {e}", err=True)
            errors_ocurred = True

    if errors_ocurred:
        raise typer.Exit(code=1)


@app.command()
def search(
    text: str,
    user: Annotated[
        Optional[bool],
        typer.Option(
            "--user/--default", help="Search only user-defined/default components"
        ),
    ] = None,
) -> None:
    """Search for a name in the database."""
    text = text.upper()

    if user:
        names = list(database.user_defined())
    elif user is not None:
        names = [name for name in database if not database.is_user_defined(name)]
    else:
        names = list(database)

    match_indices = [name.find(text) for name in names]

    for name, index in zip(names, match_indices):
        if index != -1:
            typer.echo(
                name[0:index]
                + typer.style(name[index : index + len(text)], bold=True)
                + name[index + len(text) :]
            )


def version_callback(show: bool) -> None:
    if show:
        typer.echo(f"{__package__} {__version__}")
        raise typer.Exit


@app.callback()
def common(
    version: Annotated[
        bool,
        typer.Option(
            "--version", help="Show version and exit.", callback=version_callback
        ),
    ] = False,
) -> None:
    """Database of electrolytes and their properties."""


if __name__ == "__main__":
    app(prog_name=__package__)



import contextlib

import pytest
from typer.testing import CliRunner

import electrolytes
from electrolytes import Constituent, database
from electrolytes.__main__ import app

runner = CliRunner()


def test_version() -> None:
    result = runner.invoke(app, ["--version"])
    assert result.exit_code == 0
    assert electrolytes.__version__ in result.stdout


def test_ls() -> None:
    assert "SILVER" in database
    assert not database.is_user_defined("SILVER")

    result = runner.invoke(app, ["ls"])
    assert result.exit_code == 0
    assert "SILVER" in result.stdout

    result = runner.invoke(app, ["ls", "--default"])
    assert result.exit_code == 0
    assert "SILVER" in result.stdout

    result = runner.invoke(app, ["ls", "--user"])
    assert result.exit_code == 0
    assert "SILVER" not in result.stdout


def test_no_rm_default() -> None:
    assert "SILVER" in database
    assert not database.is_user_defined("SILVER")

    result = runner.invoke(app, ["rm", "SILVER"])
    assert result.exit_code != 0

    result = runner.invoke(app, ["rm", "-f", "SILVER"])
    assert result.exit_code != 0


def test_info() -> None:
    result = runner.invoke(app, ["info"])
    assert result.exit_code == 0
    assert str(len(database)) in result.stdout

    result = runner.invoke(app, ["info", "SILVER", "ZINC"])
    assert result.exit_code == 0
    assert "SILVER" in result.stdout
    assert "ZINC" in result.stdout
    assert "user-defined" not in result.stdout


def test_add_and_rm() -> None:
    name = "TesT2322745845"
    with contextlib.suppress(KeyError):
        del database[name]

    assert name not in database
    with pytest.raises(KeyError):
        database[name]

    result = runner.invoke(app, ["ls"])
    assert result.exit_code == 0
    assert name.upper() not in result.stdout

    result = runner.invoke(app, ["info", name])
    assert result.exit_code != 0

    assert "SILVER" in database
    result = runner.invoke(app, ["info", "SILVER", name])
    assert result.exit_code != 0

    result = runner.invoke(app, ["add", name.lower(), "-2", "4", "5"])
    assert result.exit_code != 0

    result = runner.invoke(
        app, ["add", name.lower(), "-1", "2", "3", "-2", "4", "5", "+1", "6", "-1.5"]
    )

    assert result.exit_code == 0
    assert name in database
    c = database[name]
    assert len(c.mobilities()) == 6
    assert len(c.pkas()) == 6
    assert c.mobilities() == pytest.approx([0, 0, 6e-9, 2e-9, 4e-9, 0])
    assert c.pkas()[0] == Constituent._default_pka(+3)
    assert c.pkas()[1] == Constituent._default_pka(+2)
    assert c.pkas()[2:-1] == pytest.approx([-1.5, 3, 5])
    assert c.pkas()[-1] == Constituent._default_pka(-3)

    result = runner.invoke(
        app, ["add", name.upper(), "-1", "2", "3", "-2", "4", "5", "+1", "6", "-1.5"]
    )
    assert result.exit_code != 0

    result = runner.invoke(app, ["add", "-f", name, "+1", "2", "7", "+2", "4", "5"])
    assert result.exit_code == 0
    assert database[name].pos_count == 2

    result = runner.invoke(app, ["info", name])
    assert result.exit_code == 0
    assert name.upper() in result.stdout
    assert "user-defined" in result.stdout

    result = runner.invoke(app, ["ls"])
    assert result.exit_code == 0
    assert name.upper() in result.stdout

    result = runner.invoke(app, ["ls", "--user"])
    assert result.exit_code == 0
    assert name.upper() in result.stdout

    result = runner.invoke(app, ["ls", "--default"])
    assert result.exit_code == 0
    assert name.upper() not in result.stdout

    result = runner.invoke(app, ["rm", name])
    assert result.exit_code == 0

    with pytest.raises(KeyError):
        del database[name]

    assert name not in database

    result = runner.invoke(app, ["rm", name])
    assert result.exit_code != 0

    result = runner.invoke(app, ["rm", "-f", name])
    assert result.exit_code == 0


def test_extra_charges() -> None:
    name = "TEST1328849821"
    with contextlib.suppress(KeyError):
        del database[name]

    assert name not in database
    with pytest.raises(KeyError):
        database[name]

    result = runner.invoke(
        app,
        [
            "add",
            name,
            "+1",
            "5",
            "8",
            "+2",
            "7",
            "6",
            "+3",
            "9",
            "4",
            "+4",
            "11",
            "2",
            "-1",
            "1",
            "10",
            "-2",
            "3",
            "12",
        ],
    )

    assert result.exit_code == 0
    assert name in database
    c = database[name]
    assert len(c.mobilities()) == 8
    assert len(c.pkas()) == 8
    assert c.mobilities() == pytest.approx([11e-9, 9e-9, 7e-9, 5e-9, 1e-9, 3e-9, 0, 0])
    assert c.pkas() == pytest.approx(
        [2, 4, 6, 8, 10, 12, Constituent._default_pka(-3), Constituent._default_pka(-4)]
    )

    result = runner.invoke(app, ["info", name])
    assert result.exit_code == 0

    result = runner.invoke(app, ["rm", name])
    assert result.exit_code == 0

    with pytest.raises(KeyError):
        del database[name]

    assert name not in database

    result = runner.invoke(app, ["rm", name])
    assert result.exit_code != 0



import contextlib
import subprocess
from time import sleep

from electrolytes import Constituent, database


def test_lock_api_cli() -> None:
    name = "TES7342982891"
    with contextlib.suppress(KeyError):
        del database[name]

    assert name not in database

    database.add(Constituent(name=name, u_neg=[2.73], pkas_neg=[3.14]))

    with database:
        proc = subprocess.Popen(["electrolytes", "rm", name])
        sleep(0.1)
        del database[name]
    proc.wait()

    assert proc.returncode != 0
    assert name not in database
